// by Proteus  133575351 lawnmower-man@mail.ru

#define BITS 8	//    
#define MAX   ((1<<BITS)-1)
#define EPS 1e-6
#define PI 3,1415926535897932384626433832795
#define POW3(x) x*x*x

#define MAX2      (MAX/2)
#define MAX3      (MAX/3)
#define MAX6      (MAX/6)
#define MAX23     (MAX*2/3)

#define RX  95.047     // Observer = 2, Illuminant = D65(Daylight)
#define RY 100.000
#define RZ 108.883


#include<math.h>
#include<stdio.h>
#include<assert.h>

struct Lab {double l;double a;double b;};
struct XYZ {double x;double y;double z;};
struct RGB {int r;int g;int b;};
struct HSL {int h;int s;int l;};
struct CMYK{int c;int m;int y;int k;};

const class ColorSpace{
private:
	int Hue_to_RGB(int,int,int hue);
protected:
	int max(int x,int y) {return x<y?y:x;}
	int min(int x,int y) {return x<y?x:y;}
	int max(int x,int y,int z) {return max(x,max(y,z));}
	int min(int x,int y,int z) {return min(x,min(y,z));}
	void RotBound(int&);
	void CutBound(int&);
	void RGB_to_XYZ(RGB,XYZ&);
	void XYZ_to_RGB(XYZ,RGB&);
public:
	void RGB_to_HSL(RGB,HSL&);
	void HSL_to_RGB(HSL,RGB&);
	void RGB_to_Lab_Cie(RGB,Lab&);
	void Lab_to_RGB_Cie(Lab,RGB&);
	void RGB_to_CYMK(RGB,CMYK&);
	void CYMK_to_RGB(CMYK,RGB&);
};


void ColorSpace::RotBound(int &x)
{
#if 0
	if(x<0  ) x+=MAX;
	if(x>MAX) x-=MAX;
#else
	assert(x+MAX>=0);
	x=(x+MAX)%MAX;
#endif
}


void ColorSpace::CutBound(int &x)
{
	if(x<0  ) x=0;
	if(x>MAX) x=MAX;
}


int ColorSpace::Hue_to_RGB(int n1,int n2,int hue)
{
	RotBound(hue);
	int c;
    if     (hue<MAX6 ) c=n1+(n2-n1)*hue/MAX6;
    else if(hue<MAX2 ) c=n2;
    else if(hue<MAX23) c=n1+(n2-n1)*(MAX23-hue)/MAX6;
    else c=n1;
	CutBound(c);
	return c;
}


void ColorSpace::RGB_to_XYZ(RGB in,XYZ &out)
{
#define RGB_XYZ_c(x) if(x>0.04045) x=pow((x+0.055)/1.055,2.4); else x/=12.92;
	double r=(double)in.r/MAX;
	double g=(double)in.g/MAX;
	double b=(double)in.b/MAX;
	RGB_XYZ_c(r);
	RGB_XYZ_c(g);
	RGB_XYZ_c(b);
	out.x=r*0.4124*100+g*0.3576*100+b*0.1805*100;	
	out.y=r*0.2126*100+g*0.7152*100+b*0.0722*100;
	out.z=r*0.0193*100+g*0.1192*100+b*0.9505*100;
}

void ColorSpace::XYZ_to_RGB(XYZ in,RGB& out)
{
#define XYZ_RGB_c(x) if(x>0.0031308) x=1.055*pow(x,1/2.4)-0.055; else x*=12.92
	double r=in.x* 3.2406/100+in.y*-1.5372/100+in.z*-0.4986/100;
	double g=in.x*-0.9689/100+in.y* 1.8758/100+in.z* 0.0415/100;
	double b=in.x* 0.0557/100+in.y*-0.2040/100+in.z* 1.0570/100;
	XYZ_RGB_c(r);
	XYZ_RGB_c(g);
	XYZ_RGB_c(b);
	out.r=int(r*MAX);
	out.g=int(g*MAX);
	out.b=int(b*MAX);
	CutBound(out.r);
	CutBound(out.g);
	CutBound(out.b);
}


void ColorSpace::RGB_to_HSL(RGB in,HSL& out)
{
	int max=ColorSpace::max(in.r,in.g,in.b);
    int min=ColorSpace::min(in.r,in.g,in.b);
	int del=max-min;
	int sum=max+min;
    out.l=sum/2;
    if(!del) out.s=out.h=0; else {
		out.s=(del*MAX+sum/2)/sum;
		if      (in.r==max) out.h=      (in.g-in.b)*MAX6/del;
		else if (in.g==max) out.h=MAX3 +(in.b-in.r)*MAX6/del;
		else if (in.b==max) out.h=MAX23+(in.r-in.g)*MAX6/del;
		else assert(0);
		CutBound(out.s);
		RotBound(out.h);
	}
	assert(out.l>=0 && out.l<=MAX);
}

void ColorSpace::HSL_to_RGB(HSL in,RGB& out)
{
	int n1, n2;
	if(!in.s) out.r=out.g=out.b=in.l; else {
		n2=(MAX+in.s)*in.l/MAX;
		n1=2*in.l-n2;
		out.r=Hue_to_RGB(n1,n2,in.h+MAX3); 
		out.g=Hue_to_RGB(n1,n2,in.h     );
		out.b=Hue_to_RGB(n1,n2,in.h-MAX3);
	}
}



void ColorSpace::RGB_to_Lab_Cie(RGB in,Lab &out)
{
#define RGB_LAB_c(x,r) if(x>0.008856) x=pow(x/r,1.0/3);else x=x*7.787/r+16.0/116
	XYZ xyz;
	RGB_to_XYZ(in,xyz);
	double x=xyz.x;         
	double y=xyz.y;         
	double z=xyz.z;
	RGB_LAB_c(x,RX);
	RGB_LAB_c(y,RY);
	RGB_LAB_c(z,RZ);
	out.l= y   *116-16;
	out.a=(x-y)*500;
	out.b=(y-z)*200;
}

void ColorSpace::Lab_to_RGB_Cie(Lab in,RGB &out)
{
#define LAB_RGB_c(x,r) if(POW3(x)>0.008856) x=POW3(x)*r; else x=(x-16.0/116)/7.787*r
	XYZ xyz;
	xyz.y=(in.l+16)/116;
	xyz.x=(in.a/500)+xyz.y;
	xyz.z=xyz.y-in.b/200;
	LAB_RGB_c(xyz.x,RX);
	LAB_RGB_c(xyz.y,RY);
	LAB_RGB_c(xyz.z,RZ);
	XYZ_to_RGB(xyz,out);
}



void ColorSpace::RGB_to_CYMK(RGB in,CMYK &out)
{
	out.k=max(in.r,in.g,in.b);	
	if(!out.k) out.c=out.m=out.y=0;
	else {
		out.c=(out.k-in.r)*MAX/out.k;
		out.m=(out.k-in.g)*MAX/out.k;
		out.y=(out.k-in.b)*MAX/out.k;
	}
	out.k=MAX-out.k;
}


void ColorSpace::CYMK_to_RGB(CMYK in,RGB &out)
{
	int n=MAX-in.k;
	out.r=MAX-in.c*n/MAX-in.k;
	out.g=MAX-in.m*n/MAX-in.k;
	out.b=MAX-in.y*n/MAX-in.k;
}



void main()
{
	RGB rgb;
	XYZ xyz; 
	HSL hsl;
	Lab lab;
	CMYK cmyk;
	ColorSpace sp;
	
	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_XYZ(rgb,xyz);
	sp.XYZ_to_RGB(xyz,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_HSL(rgb,hsl);
	sp.HSL_to_RGB(hsl,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>2) _asm int 3


	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_Lab_Cie(rgb,lab);
	sp.Lab_to_RGB_Cie(lab,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>1) _asm int 3

	rgb.r=123; rgb.g=45; rgb.b=67;
	sp.RGB_to_CYMK(rgb,cmyk);
	sp.CYMK_to_RGB(cmyk,rgb);
	if(abs(rgb.r+rgb.g+rgb.b-123-45-67)>2) _asm int 3
}



